/*****************************************************************************
 * Copyright (c) 2020, rrcSmall FM10K-Documentation Contributors
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the names of its contributors
 *       may be used to endorse or promote products derived from this software
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *****************************************************************************/

#pragma once

#include <string>
#include <utility>
#include <vector>
#include <memory>
#include <algorithm>

class Configuration {
public:

    Configuration(){

    }

    class ConfigurationNode {
    public:

        enum class Type {
            Node = 0,
            ValueText,
            ValueInt,
            ValueBool,
        } type;

        ConfigurationNode(std::string name, ConfigurationNode::Type type = ConfigurationNode::Type::Node) : name(std::move(name)), type(type){

        }

        std::string name;
        std::vector<ConfigurationNode> possibleNodes;
        std::string value;

        std::string getTypeString() const{
            switch (type) {
                case Type::Node:
                    return "";
                case Type::ValueText:
                    return "text";
                case Type::ValueInt:
                    return "int";
                case Type::ValueBool:
                    return "bool";
            }
        }

        const ConfigurationNode* getSubNode(const std::string& k) const{
            for (auto& node : possibleNodes){
                if(node.name == k){
                    return &node;
                }
            }

            return nullptr;
        }

        ConfigurationNode& getOrCreateSubNode(const std::string& k){
            for (auto& node : possibleNodes){
                if(node.name == k){
                    return node;
                }
            }

            possibleNodes.emplace_back(ConfigurationNode(k, ConfigurationNode::Type::Node));
            return possibleNodes[possibleNodes.size() - 1];
        }

        bool getBool() const{
            return value == "true";
        }

        int getInteger() const{
            return std::stoi(value, nullptr, 0);
        }

        const std::string& getText() const{
            return value;
        }
    };

    ConfigurationNode mainNode = ConfigurationNode("");

    std::vector<std::string> getAllEntries() const{
        std::vector<std::string> entries;

        for(const auto& n : mainNode.possibleNodes){
            if(n.type == ConfigurationNode::Type::Node){
                getAllEntriesForNode(entries, n, n.name + ".");
            }
        }

        std::sort(entries.begin(), entries.end());

        return entries;
    }

    void getAllEntriesForNode(std::vector<std::string>& entries, const ConfigurationNode& node, const std::string& key) const{
        for(const auto& n : node.possibleNodes){
            if(n.type != ConfigurationNode::Type::Node){
                entries.push_back(key + n.name + " " + n.getTypeString() + " " + n.value);
            }else{
                getAllEntriesForNode(entries, n, key + n.name + ".");
            }
        }
    }

    const ConfigurationNode& getEntry(const std::string& entry) const{
        if(entry.empty() || entry[0] == '#'){
            return mainNode;
        }
        std::string currentKey;
        currentKey.reserve(16);

        const ConfigurationNode* currentNode = &mainNode;

        for(char c : entry){
            if((c == '.' || c == ' ' || c == '\t')){
                if(!currentKey.empty()){
                    currentNode = currentNode->getSubNode(currentKey);

                    if(currentNode == nullptr){
                        return mainNode;
                    }

                    currentKey.clear();

                    if(c == ' ' || c == '\t'){
                        return *currentNode;
                    }
                }
            }else{
                currentKey += c;
            }
        }

        return mainNode;
    }

    bool addEntry(const std::string& entry){
        if(entry.empty() || entry[0] == '#'){
            return false;
        }
        std::string currentKey;
        currentKey.reserve(16);

        uint32_t state = 0;

        ConfigurationNode* currentNode = &mainNode;

        for(size_t i = 0; i < entry.size(); ++i){
            auto c = entry[i];
            if(state == 0){
                if((c == '.' || c == ' ' || c == '\t')){
                    if(!currentKey.empty()){
                        currentNode = &currentNode->getOrCreateSubNode(currentKey);

                        currentKey.clear();

                        if(c == ' ' || c == '\t'){
                            state = 1;
                        }
                    }
                }else{
                    currentKey += c;
                }
            } else if(state == 1){
                if((c == ' ' || c == '\t')){
                    if(!currentKey.empty()){
                        if(currentKey == "int"){
                            currentNode->type = ConfigurationNode::Type::ValueInt;
                        }else if(currentKey == "text"){
                            currentNode->type = ConfigurationNode::Type::ValueText;
                        }else if(currentKey == "bool"){
                            currentNode->type = ConfigurationNode::Type::ValueBool;
                        }

                        currentKey.clear();

                        state = 2;
                    }
                }else{
                    currentKey += c;
                }
            } else if(state == 2){
                if(i == (entry.size() - 1)){
                    currentKey += c;
                    currentNode->value = currentKey;
                }else if((c == ' ' || c == '\t') && currentKey.empty()){

                } else {
                    currentKey += c;
                }
            }
        }

        return state == 2;
    }

};