/*****************************************************************************
 * Copyright (c) 2020, rrcSmall FM10K-Documentation Contributors
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the names of its contributors
 *       may be used to endorse or promote products derived from this software
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *****************************************************************************/

#include "Copy.h"
#include "../Registers.h"
#include <iostream>
#include <iomanip>

void Instruction::Copy::fromBytes(uint32_t offset, const std::vector<uint8_t> &bytes) {
    _address = offset;

    uint32_t command = bytes[offset++];

    count = (1 + ((command >> 4) & 0b11));
    addressOffsetBB = (command >> 2) & 0b11;
    addressOffsetAA = command & 0b11;

    addressAA = bytes[offset++] << 16;
    addressAA |= bytes[offset++] << 8;
    addressAA |= bytes[offset++];

    offset++;

    addressBB = bytes[offset++] << 16;
    addressBB |= bytes[offset++] << 8;
    addressBB |= bytes[offset++];



    _endAddress = offset;
}

std::string Instruction::Copy::toString() const {
    std::stringstream op;
    op << "COPY " << std::dec << (uint32_t) count  << " WORDS FROM " << getAddressRegisterName(addressAA, addressOffsetAA) << " TO " << getAddressRegisterName(addressBB, addressOffsetBB);
    return op.str();
}

std::vector<uint32_t> Instruction::Copy::getPossibleBranches() const {
    return std::vector<uint32_t>{_endAddress};
}

std::vector<std::pair<uint32_t, std::unordered_map<uint32_t, uint32_t>>> Instruction::Copy::execute(AnalysisState &state) const {

    uint32_t memoryOffsetAA = state.getAddressOffset(addressOffsetAA);
    uint32_t memoryOffsetBB = state.getAddressOffset(addressOffsetBB);

    for(uint32_t i = 0; i < count; ++i){
        state.setRegister(memoryOffsetAA + addressOffsetAA + i, state.getRegister(memoryOffsetBB + addressBB));
    }

    state.current = _endAddress;

    return std::vector<std::pair<uint32_t, std::unordered_map<uint32_t, uint32_t>>>();
}


std::vector<uint8_t> Instruction::Copy::toBytes() const {
    std::vector<uint8_t> bytes;

    bytes.emplace_back((uint8_t)getCommand() | ((count - 1) << 4) | (addressOffsetBB << 2) | addressOffsetAA);

    bytes.emplace_back((addressAA >> 16) & 0xFF);
    bytes.emplace_back((addressAA >> 8) & 0xFF);
    bytes.emplace_back(addressAA & 0xFF);

    bytes.emplace_back(0);

    bytes.emplace_back((addressBB >> 16) & 0xFF);
    bytes.emplace_back((addressBB >> 8) & 0xFF);
    bytes.emplace_back(addressBB & 0xFF);

    return bytes;
}
